#!/usr/bin/python3 -B
import pathlib
import sys

# Replaced during install; must be a single line
# fmt: off
DEBPUTY_ROOT_DIR = pathlib.Path(__file__).parent / 'src'
# fmt: on

if __name__ == "__main__":
    # setup PYTHONPATH: add our installation directory.
    sys.path.insert(0, str(DEBPUTY_ROOT_DIR))
    from debputy.commands.deb_materialization import main

    main()
